// Copyright © Tavian Barnes <tavianator@tavianator.com>
// SPDX-License-Identifier: 0BSD

#include "bfs.h"

const char bfs_version[] = {
#include "version.i"
};

const char bfs_confflags[] = {
#include "confflags.i"
};

const char bfs_cc[] = {
#include "cc.i"
};

const char bfs_cppflags[] = {
#include "cppflags.i"
};

const char bfs_cflags[] = {
#include "cflags.i"
};

const char bfs_ldflags[] = {
#include "ldflags.i"
};

const char bfs_ldlibs[] = {
#include "ldlibs.i"
};
