// Copyright © Tavian Barnes <tavianator@tavianator.com>
// SPDX-License-Identifier: 0BSD

/// _CFLAGS += -std=c23
/// ---
/// _CFLAGS += -std=c2x
/// ---
/// _CFLAGS += -std=c17

int main(void) {
	return 0;
}
